#include <fcntl.h>
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <unistd.h>
#include <time.h>
#include "gp2x.h"

struct gp2x_init
{
	char *regs;
	int fd_mem;
	int page_size;

	gp2x_init();
	~gp2x_init();
};

static gp2x_init static_init;

static void sound_destroy();

static inline unsigned short input2(int address)
{
	return *(volatile unsigned short *) (static_init.regs + address);
}

static inline unsigned int input4(int address)
{
	return *(volatile unsigned int *) (static_init.regs + address);
}

static inline void output2(int address, unsigned short value)
{
	*(volatile unsigned short *) (static_init.regs + address) = value;
}

static inline void output4(int address, unsigned value)
{
	*(volatile unsigned int *) (static_init.regs + address) = value;
}

gp2x_init::gp2x_init()
{
	fd_mem = open("/dev/mem", O_RDWR);

	regs = (char *) mmap(0, 0x10000, PROT_READ | PROT_WRITE, MAP_SHARED, fd_mem, 0xc0000000);

	page_size = sysconf(_SC_PAGESIZE);
}

gp2x_init::~gp2x_init()
{
	sound_destroy();

	munmap(regs, 0x10000);
	close(fd_mem);

	chdir("/usr/gp2x");
	execl("gp2xmenu", "gp2xmenu", NULL);
}


// joystick

int gp2x::joystick::read_state()
{
	int value = input2(0x1198) & 0x00FF;

	if (value == 0xFD) value=0xFA;
	if (value == 0xF7) value=0xEB;
	if (value == 0xDF) value=0xAF;
	if (value == 0x7F) value=0xBE;

	return BUTTON_ALL & ~((input2(0x1184) & 0xFF00) | value | (input2(0x1186) << 16));
}

static void sleep_ns(int ns)
{
	struct timespec ts;
	ts.tv_sec = 0;
	ts.tv_nsec = ns;

	nanosleep(&ts, NULL);
}

int gp2x::joystick::wait_button_press()
{
	while (read_state() != 0) sleep_ns(1000000);

	int ret = read_state();
	while (ret == 0)
	{
		sleep_ns(1000000);
		ret = read_state();
	};

	while (read_state() != 0) sleep_ns(1000000);

	return ret;
}


// timer

void gp2x::timer::pause(unsigned int ticks)
{
	unsigned int target = input4(0x0A00) + ticks;
	while (input4(0x0A00) < target);
}

unsigned int gp2x::timer::read_ticks()
{
	return input4(0x0A00);
}


// video

static unsigned char *m_map_next = (unsigned char *) 0x03000000;
static int m_screen_scale_width = -1;
static int m_screen_scale_height = -1;
static bool m_palette_inited = false;
static int m_bpp = -1;

void gp2x::video::wait_v_sync()
{
	while (input2(0x1182) & 0x10);
}

void gp2x::video::wait_h_sync()
{
	while (input2(0x1182) & 0x20);
}


static void set_screen_scale(int width, int height, int pitch)
{
	if (m_screen_scale_width == width && m_screen_scale_height == height) return;

	m_screen_scale_width = width;
	m_screen_scale_height = height;

	//int bpp = (input2(0x28DA) >> 9) & 0x3;

	//output2(0x2906, (unsigned short) (1024.0 * (width / 320.0)));
	//output4(0x2908, (unsigned int) (320.0 * bpp * (height / 240.0)));

	output2(0x2906, (width << 4) / 5);
	output4(0x2908, (pitch * height) / 240);

	//int scale_y = width * height / 240;
	//int scale_y = ((352+64) * height) / 240;
	//output2(0x2908, scale_y & 0xffff);
	//output2(0x290a, scale_y >> 16);

}

static void init_palette()
{
	const unsigned char points_3[8] = { 0, 36, 73, 109, 146, 182, 219, 255 };
	const unsigned char points_2[4] = { 0, 85, 170, 255 };

	m_palette_inited = true;

	output2(0x2958, 0);

	for (int i = 0; i < 256; i++)
	{
		output2(0x295A, (points_3[(i & 0xe0) >> 5] << 8) | points_2[i & 0x03]);
		output2(0x295A, points_3[(i & 0x1c) >> 2]);
	}
}

static void set_bpp(int bpp)
{
	if (m_bpp == bpp) return;
	m_bpp = bpp;

	output2(0x28DA, (((bpp + 1) / 8) << 9) | 0x0AB);
}

// video - surface8

static void set_framebuffer(unsigned int address, int pitch)
{
	output2(0x290C, pitch);

	output2(0x290E, (unsigned short) (address & 0xffff));
	output2(0x2912, (unsigned short) (address & 0xffff));

	output2(0x2910, (unsigned short) (address >> 16));
	output2(0x2914, (unsigned short) (address >> 16));
}

gp2x::video::surface8::surface8(int width, int height, bool hardware, bool double_buffered)
{
	if (hardware && !m_palette_inited) init_palette();

	m_double_buffered = hardware && double_buffered;
	m_first_buffer = true;

	m_width = width;
	m_height = height;

	m_size1 = height * width;
	if (hardware) m_size1 = (m_size1 + static_init.page_size - 1) & ~(static_init.page_size - 1);
	else m_size1 = (m_size1 + 7) & ~7U;

	m_size = m_size1 + (m_double_buffered ? m_size1 : 0);

	if (hardware)
	{
		m_physical = m_map_next, m_map_next += m_size;
		m_virtual = (unsigned char *) mmap(0, m_size, PROT_READ | PROT_WRITE, MAP_SHARED, static_init.fd_mem, (__off_t) m_physical);
	}
	else
	{
		m_physical = 0;
		m_virtual = (unsigned char *) malloc(m_size);
	}

	memset(m_virtual, 0, m_size);

	m_data = m_virtual;
}

gp2x::video::surface8::~surface8()
{
	if (m_physical) munmap(m_virtual, m_size);
	else free(m_virtual);
}

void gp2x::video::surface8::flip()
{
	flip(0, 0, 320, 240);
}

void gp2x::video::surface8::flip(int x, int y)
{
	flip(x, y, 320, 240);
}

void gp2x::video::surface8::flip(int x, int y, int w, int h) // scale to w x h
{
	if (!m_physical) return;
	if (m_screen_scale_width != w || m_screen_scale_height != h) set_screen_scale(w, h, m_width);

	if (m_bpp != 8) set_bpp(8);

	unsigned address = unsigned(m_physical + m_width * y + x);

	if (m_double_buffered)
	{
		if (!m_first_buffer) address += m_size1;
		m_first_buffer = !m_first_buffer;
		m_data = m_first_buffer ? m_virtual : m_virtual + m_size1;
	}

	set_framebuffer(address, m_width);
}

unsigned char gp2x::video::surface8::color(int r, int g, int b)
{
	return ((r >> 3) & 0x1c) | (g & 0xe0) | ((b >> 6) & 0x03);
}

void gp2x::video::surface8::rect(int x, int y, int w, int h, unsigned char color)
{
	register unsigned char *dest = m_data;
	dest += m_width * y + x;

	for (int i = 0; i < h; i++)
	{
		memset(dest, color, w);
		dest += m_width;
	}
}

int gp2x::video::surface8::text_width(const unsigned char *font, const char *text, bool fixed_width)
{
	int fw = font[0];

	if (fixed_width) return strlen(text) * fw;

	int tw = 0;
	int fh = font[1];

	for (int i = 0; text[i]; i++)
	{
		unsigned char c = text[i];
		if (c < 33 || c > 126 && c < 161) tw += fw;
		else
		{
			const unsigned char *src = c < 161 ? &font[2 + (fh + 1) * (c - 33)] : &font[2 + (fh + 1) * (c - 161)];
			tw += 1 + (src[0] & 0x0f);
		}
	}

	return tw;
}

static int print_char(const unsigned char *font, unsigned char *dest, unsigned char c, unsigned char color_fg, unsigned char color_bg, bool fixed_width, int pitch)
{
	int fw = font[0], fh = font[1];

	if (c < 33 || c > 126 && c < 161) return fw;

	const unsigned char *src = c < 161 ? &font[2 + (fh + 1) * (c - 33)] : &font[2 + (fh + 1) * (c - 161)];
	int shift = src[0] >> 4;
	int cw = fixed_width ? fw : (src[0] & 0x0f);
	bool bg = (color_fg != color_bg);

	for (int h = 1; h <= fh; h++)
	{
		unsigned char d = src[h];
		if (!fixed_width && shift > 0) d <<= shift;
		for (int w = 0; w < cw; w++, d <<= 1) if (d & 0x80) dest[w] = color_fg; else if (bg) dest[w] = color_bg;
		dest += pitch;
	}

	return fixed_width ? fw : cw + 1;
}

int gp2x::video::surface8::print(const unsigned char *font, int x, int y, const char *text, unsigned char color_fg, unsigned char color_bg, bool fixed_width)
{
	int tw = 0;

	unsigned char *dest = m_data;
	dest += m_width * y + x;

	unsigned char *dest_limit = dest + 320 - x - font[0];

	for (int i = 0; text[i] && dest < dest_limit; i++)
	{
		int cw = print_char(font, dest, (unsigned char) text[i], color_fg, color_bg, fixed_width, m_width);
		dest += cw;
		tw += cw;
	}

	return tw;
}

void gp2x::video::surface8::clear(bool both)
{
	if (both) memset(m_virtual, 0, m_size);
	else memset(m_data, 0, m_size1);
}

int gp2x::video::sync_fps(unsigned fps)
{
	static unsigned int t_prev = 0, t0 = 0;
	static int frames = -1;
	static unsigned fps_prev = 0, delta;
	static int fps_current = 0;

	if (fps_prev != fps) delta = gp2x::timer::HZ / fps, fps_prev = fps;

	unsigned t = gp2x::timer::read_ticks();

	if (frames >= 0)
	{
		unsigned t_target = t_prev + delta;
		bool done = false;

		do
		{
			t = gp2x::timer::read_ticks();
			done = (t < t_prev || t >= t_target);
		}
		while (!done);
	}
	else
	{
		frames = 0;
		t0 = t;
	}

	frames++;

	t_prev = t;

	if ((t - t0) >= gp2x::timer::HZ) fps_current = frames, frames = 0, t0 = t;

	return fps_current;
}

void gp2x::video::surface8::blit(surface8 *source, int sx, int sy, int sw, int sh, int dx, int dy)
{
	register unsigned char *s = source->data() + source->width() * sy + sx;
	register unsigned char *d = data() + width() * dy + dx;

	for (register int i = 0; i < sh; i++)
	{
		memcpy(d, s, sw);
		s += source->width();
		d += width();
	}
}


// sound

static int fd_mixer = 0, fd_dsp = 0;
static bool sound_paused = true;
static gp2x::sound::sound_producer_func sound_producer = 0;
static unsigned short *sound_buffer = 0;
static int sound_buffer_samples = -1;
static int sound_buffer_bytes = -1;
static bool sound_do = true;
static pthread_t sound_thread;
static pthread_mutex_t sound_mutex, sound_mutex_pause;

extern "C"
static void *sound_thread_player(void *)
{
	while (sound_do)
	{
		pthread_mutex_lock(&sound_mutex_pause);
		pthread_mutex_lock(&sound_mutex);

		sound_producer(sound_buffer, sound_buffer_samples);

		pthread_mutex_unlock(&sound_mutex);
		pthread_mutex_unlock(&sound_mutex_pause);

		write(fd_dsp, sound_buffer, sound_buffer_bytes);
	}

	return 0;
}


void gp2x::sound::setup(int sound_freq, bool stereo, sound_producer_func producer)
{
	if (!fd_dsp)
	{
		fd_dsp = open("/dev/dsp", O_WRONLY);
		fd_mixer = open("/dev/mixer", O_RDWR);

		pthread_mutex_init(&sound_mutex, 0);
		pthread_mutex_init(&sound_mutex_pause, 0);

		pthread_mutex_lock(&sound_mutex_pause);
		sound_paused = true;
		pthread_create(&sound_thread, 0, sound_thread_player, 0);
	}

	int frag = 0x80009;
	ioctl(fd_dsp, SNDCTL_DSP_SETFRAGMENT, &frag);

	pause();

	const int bits = 16;
	int stereo_int = stereo ? 1 : 0;

	ioctl(fd_dsp, SNDCTL_DSP_SETFMT, &bits);
	ioctl(fd_dsp, SNDCTL_DSP_STEREO, &stereo_int);
	ioctl(fd_dsp, SNDCTL_DSP_SPEED, &sound_freq);

	sound_producer = producer;

	audio_buf_info abi;
	ioctl(fd_dsp, SNDCTL_DSP_GETOSPACE, &abi);

	if (sound_buffer_samples != abi.fragsize)
	{
		sound_buffer_samples = abi.fragsize;
		sound_buffer_bytes = sound_buffer_samples << 1;
		if (sound_buffer) delete[] sound_buffer;
		sound_buffer = new unsigned short[sound_buffer_samples];
		memset(sound_buffer, 0, sound_buffer_bytes);
	}

	volume(100, 100);
}

void gp2x::sound::play()
{
	if (!sound_paused) return;
	pthread_mutex_unlock(&sound_mutex_pause);
	sound_paused = false;
}

void gp2x::sound::pause()
{
	if (sound_paused) return;
	pthread_mutex_lock(&sound_mutex_pause);
	sound_paused = true;
}

void gp2x::sound::lock()
{
	pthread_mutex_lock(&sound_mutex);
}

void gp2x::sound::unlock()
{
	pthread_mutex_unlock(&sound_mutex);
}

void gp2x::sound::volume(int l, int r)
{
	l = (l << 2) / 5;
	r = (r << 2) / 5;
	int v = (l << 8) | r;

	ioctl(fd_mixer, SOUND_MIXER_WRITE_PCM, &v);
}

void sound_destroy()
{
	if (!fd_dsp) return;

	pthread_mutex_unlock(&sound_mutex);
	pthread_mutex_unlock(&sound_mutex_pause);

	sound_do = false;

	close(fd_mixer);
	close(fd_dsp);

	pthread_join(sound_thread, 0);
	pthread_mutex_destroy(&sound_mutex);
	pthread_mutex_destroy(&sound_mutex_pause);
}
