#include <string.h>
#include "types.h"
#include "io.h"
#include "cpu.h"
#include "gp2x.h"
#include "sound.h"

static int sound_pos, sound_samples;
static sword sound_buffer[SOUND_BUF_SIZE];

#define NOISE_ON	0x80
#define VOICE_ON	0x80
#define DDA_ON		0x40
#define LFO_OFF		0x80
#define SBASE_CLOCK	3580000

#define MAX_FREQ	(SOUND_FREQ * 4)

#define SPEED_SHIFT	12
#define DEFAULT_SPEED	((SOUND_FREQ << 12) / BASE_CLOCK)

#define VD		65536
#define NPM		9
#define NNM		-10

typedef void (*func_generator)(int ch, sword *dest, int count);

struct sound_vars
{
	byte reg_v, reg_vol, reg_lfo_control, reg_lfo_freq;
	byte reg_noise[6], reg_status[6], reg_balance[6];
	pair reg_freq[6];
	sbyte reg_samples[6][32];
	byte reg_samples_pos[6];

	byte change_samples[6];

	int ch_vol[6];

	int noise_delta[6];
	sword noise_sample[6], noise_sample_xor[6];
	dword noise_random[6];
	int noise_freq[6];

	short voice_samples[6][32];
	long voice_delta[6], voice_freq[6];
	byte reg_voice_pos[6];

	sword dda_sample[6];
	byte reg_dda_value[6];

	bool sound_on;
	sword *mix_dest;
	dword last_cpu;

	func_generator generator[6];
};

static sound_vars snd;

static void update_samples(int ch)
{
	int vol = snd.ch_vol[ch];

	for (int i = 0; i < 32; i++) snd.voice_samples[ch][i] = (snd.reg_samples[ch][i] * vol) >> 16;
	snd.change_samples[ch] = 0;
}

static void generator_silence(int, sword *, int)
{
}

static void generator_noise(int ch, sword *dest, int count)
{
	int freq, delta;
	dword random;
	sword sample;

	freq = snd.noise_freq[ch];
	if (!freq) return;

	delta = snd.noise_delta[ch];
	sample = snd.noise_sample[ch];
	random = snd.noise_random[ch];

	for (int i = count; i > 0; i--)
	{
		*dest++	+= sample;

		delta -= freq;
		while (delta < 0)
		{
			dword x;

			delta += SOUND_FREQ;

			random = (random & 0x80000000) | (random >>	1);
			x =	random & 0x40;
			if (x) sample ^= snd.noise_sample_xor[ch];
			random ^= x	<< 25;
		}
	}

	snd.noise_sample[ch] = sample;
	snd.noise_random[ch] = random;
	snd.noise_delta[ch] = delta;
}

static void generator_dda(int ch, sword *dest, int count)
{
	sword sample = snd.dda_sample[ch];

	for (int i = count; i > 0; i--) *dest++ += sample;
}

static void generator_voice(int ch, sword *dest, int count)
{
	byte pos;
	int freq, delta;
	sword sample;

	freq = snd.voice_freq[ch];
	if (!freq) return;

	if (snd.change_samples[ch]) update_samples(ch);

	delta = snd.voice_delta[ch];
	sample = snd.voice_samples[ch][snd.reg_voice_pos[ch]];
	pos = snd.reg_voice_pos[ch];

	for (int i = count; i > 0; i--)
	{
		*dest++ += sample;
		delta -= freq;
		while (delta < 0)
		{
			delta += SOUND_FREQ;
			pos = (pos + 1) & 0x1f;
			sample = snd.voice_samples[ch][pos];
		}
	}

	snd.voice_delta[ch] = delta;
	snd.reg_voice_pos[ch] = pos;
}

static inline int min(int a, int b)
{
	return a <= b ? a : b;
}

static void sound_mix(int count, bool lock)
{
	if (lock) gp2x::sound::lock();

	while (count > 0 && sound_samples < SOUND_BUF_SIZE)
	{
		int start_pos = (sound_pos + sound_samples) & (SOUND_BUF_SIZE - 1);
		int rcount = min(count, min(SOUND_BUF_SIZE - sound_samples, SOUND_BUF_SIZE - start_pos));

		if (rcount <= 0) return;

		sword *dest = &sound_buffer[start_pos];

		for (int i = rcount - 1; i >= 0; i--) dest[i] = 0;
  		for (int i = 0; i < 6; i++) snd.generator[i](i, dest, rcount);

		sound_samples += rcount;
		count -= rcount;
	}

	if (lock) gp2x::sound::unlock();
}

void sound_mix()
{
	gp2x::sound::lock();

	if (!snd.last_cpu)
	{
		snd.last_cpu = cpu_cycles_counter;
	}
	else
	{
		int count = ((cpu_cycles_counter - snd.last_cpu) * DEFAULT_SPEED) >> SPEED_SHIFT;
		if (count > 0 && sound_samples < SOUND_BUF_SIZE)
		{
			snd.last_cpu = cpu_cycles_counter;
			sound_mix(count, false);
		}
	}

	gp2x::sound::unlock();
}

void sound_producer(word *buffer, int samples)
{
	while (samples > 0)
	{
		if (sound_samples < samples) sound_mix(samples - sound_samples, false);

		int count = min(min(SOUND_BUF_SIZE - sound_pos, sound_samples), samples);

		if (count > 0)
		{
			memcpy(buffer, &sound_buffer[sound_pos], count << 1);
			buffer += count;
			samples -= count;
			sound_samples -= count;
			sound_pos = (sound_pos + count) & (SOUND_BUF_SIZE - 1);
		}
	}
}

byte IO_read_0800()
{
	return snd.reg_v;
}

byte IO_read_0801()
{
	return snd.reg_vol;
}

byte IO_read_0802()
{
	return snd.reg_freq[snd.reg_v].B.l;
}

byte IO_read_0803()
{
	return snd.reg_freq[snd.reg_v].B.h;
}

byte IO_read_0804()
{
	return snd.reg_status[snd.reg_v];
}

byte IO_read_0805()
{
	return snd.reg_balance[snd.reg_v];
}

byte IO_read_0806()
{
	byte ret = snd.reg_samples[snd.reg_v][snd.reg_samples_pos[snd.reg_v]] + 16;
	snd.reg_samples_pos[snd.reg_v] = (snd.reg_samples_pos[snd.reg_v] + 1) & 0x1f;
	return ret;
}

byte IO_read_0807()
{
	return snd.reg_noise[snd.reg_v];
}

byte IO_read_0808()
{
	return snd.reg_lfo_freq;
}

byte IO_read_0809()
{
	return snd.reg_lfo_control;
}

void IO_write_0800(byte value)
{
	snd.reg_v = (value & 0x07) % 6;
}

static void update_volume(int ch)
{
#define VOL_MUL 1363
	static const long vol_tab[60+32] =
	{
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,100*VOL_MUL,
		100*VOL_MUL,  451*VOL_MUL,  508*VOL_MUL,  573*VOL_MUL,
		646*VOL_MUL,  728*VOL_MUL,  821*VOL_MUL,  925*VOL_MUL,
		1043*VOL_MUL, 1175*VOL_MUL, 1325*VOL_MUL, 1493*VOL_MUL,
		1683*VOL_MUL, 1898*VOL_MUL, 2139*VOL_MUL, 2411*VOL_MUL,
		2718*VOL_MUL, 3064*VOL_MUL, 3454*VOL_MUL, 3893*VOL_MUL,
		4388*VOL_MUL, 4947*VOL_MUL, 5576*VOL_MUL, 6285*VOL_MUL,
		7085*VOL_MUL, 7986*VOL_MUL, 9002*VOL_MUL, 10148*VOL_MUL,
		11439*VOL_MUL,12894*VOL_MUL,14535*VOL_MUL,16384*VOL_MUL
	};

	long left = vol_tab[((snd.reg_balance[ch] >> 3) & 0x1e) + (snd.reg_status[ch] & 0x1f) + ((snd.reg_vol >> 3) & 0x1e)];
	long right = vol_tab[((snd.reg_balance[ch] << 1) & 0x1e) + (snd.reg_status[ch] & 0x1f) + ((snd.reg_vol << 1) & 0x1e)];
	if (left < right) left = right;
	snd.ch_vol[ch] = left;

	snd.noise_sample[ch] = (NPM * snd.ch_vol[ch]) / VD;
	snd.noise_sample_xor[ch] = snd.noise_sample[ch] ^ ((NNM * snd.ch_vol[ch]) / VD);
	snd.dda_sample[ch] = (snd.reg_dda_value[ch] * snd.ch_vol[ch]) / VD;
	snd.change_samples[ch] = 1;
}

void IO_write_0801(byte value)
{
	if (snd.reg_vol == value) return;
	sound_mix();
	snd.reg_vol = value;
	for (int i = 0; i < 6; i++) update_volume(i);
}

void IO_write_0802(byte value)
{
	if (snd.reg_freq[snd.reg_v].B.l == value) return;
	if ((snd.reg_noise[snd.reg_v] & NOISE_ON) == 0 && (snd.reg_status[snd.reg_v] & VOICE_ON)) sound_mix();
	snd.reg_freq[snd.reg_v].B.l = value;
	if (snd.reg_freq[snd.reg_v].W == 0)
	{
		snd.voice_freq[snd.reg_v] = 0;
	}
	else
	{
		snd.voice_freq[snd.reg_v] = SBASE_CLOCK / snd.reg_freq[snd.reg_v].W;
		if (snd.voice_freq[snd.reg_v] > MAX_FREQ) snd.voice_freq[snd.reg_v] = 0;
	}
}

void IO_write_0803(byte value)
{
	value &= 0x0f;
	if (snd.reg_freq[snd.reg_v].B.h == value) return;
	if ((snd.reg_noise[snd.reg_v] & NOISE_ON) == 0 && (snd.reg_status[snd.reg_v] & VOICE_ON)) sound_mix();
	snd.reg_freq[snd.reg_v].B.h = value;
	if (snd.reg_freq[snd.reg_v].W == 0)
	{
		snd.voice_freq[snd.reg_v] = 0;
	}
	else
	{
		snd.voice_freq[snd.reg_v] = SBASE_CLOCK / snd.reg_freq[snd.reg_v].W;
		if (snd.voice_freq[snd.reg_v] > MAX_FREQ) snd.voice_freq[snd.reg_v] = 0;
	}
}
void IO_write_0804(byte value)
{
	if (snd.reg_status[snd.reg_v] == value) return;
	sound_mix();

	if ((snd.reg_status[snd.reg_v] ^ value) & 0x1f)
	{
		snd.reg_status[snd.reg_v] = value;
		update_volume(snd.reg_v);
	}

	snd.reg_status[snd.reg_v] = value;

	if ((value & VOICE_ON) == 0)
	{
		if ((snd.reg_noise[snd.reg_v] & NOISE_ON) == 0) snd.generator[snd.reg_v] = generator_silence;
		if (value & DDA_ON)
		{
			snd.reg_samples_pos[snd.reg_v] = 0;
			snd.reg_voice_pos[snd.reg_v] = 0;
			snd.voice_delta[snd.reg_v] = SOUND_FREQ;
		}
	}
	else
	{
		if ((snd.reg_noise[snd.reg_v] & NOISE_ON) == 0)
		{
			if ((value & DDA_ON) == 0) snd.generator[snd.reg_v] = generator_voice;
			else snd.generator[snd.reg_v] = generator_dda;
        }
	}
}

void IO_write_0805(byte value)
{
	if (snd.reg_balance[snd.reg_v] == value) return;

	sound_mix();
	snd.reg_balance[snd.reg_v] = value;
	update_volume(snd.reg_v);
}

void IO_write_0806(byte value)
{
	byte x;

	value &= 0x1f;
	x = snd.reg_status[snd.reg_v] & (VOICE_ON + DDA_ON);
	if (x == (DDA_ON + VOICE_ON))
	{
		sound_mix();
		snd.reg_dda_value[snd.reg_v] = value - 16;
		snd.dda_sample[snd.reg_v] = ((value - 16 ) * snd.ch_vol[snd.reg_v]) / VD;
	}
	else if (x == 0)
	{
		snd.reg_samples[snd.reg_v][snd.reg_samples_pos[snd.reg_v]] = value - 16;
		snd.reg_samples_pos[snd.reg_v] = (snd.reg_samples_pos[snd.reg_v] + 1) & 0x1f;
		snd.change_samples[snd.reg_v] = 1;
	}
}

void IO_write_0807(byte value)
{
	if (value == snd.reg_noise[snd.reg_v]) return;
	if (snd.reg_v < 4) return;

	sound_mix();

	if ((value ^ 0x1f) & 0x1f)
	{
		snd.noise_freq[snd.reg_v] = SBASE_CLOCK / (((value & 0x1f) ^ 0x1f) << 6);
		if (snd.noise_freq[snd.reg_v] > SOUND_FREQ) snd.noise_freq[snd.reg_v] = SOUND_FREQ;
	}
	else
	{
		snd.noise_freq[snd.reg_v] = 0;
	}

	if (value & NOISE_ON)
	{
		if ((snd.reg_noise[snd.reg_v] & NOISE_ON) == 0)
		{
			snd.noise_delta[snd.reg_v] = SOUND_FREQ;
			snd.generator[snd.reg_v] = generator_noise;
		}
	}
        else
        {
        	if (snd.reg_status[snd.reg_v] & VOICE_ON)
        	{
        		if (snd.reg_status[snd.reg_v] & DDA_ON) snd.generator[snd.reg_v] = generator_dda;
        		else snd.generator[snd.reg_v] = generator_voice;
        	}
        	else
        		snd.generator[snd.reg_v] = generator_silence;
        }

	snd.reg_noise[snd.reg_v] = value;
}

void IO_write_0808(byte value)
{
	snd.reg_lfo_freq = value;
}

void IO_write_0809(byte value)
{
	snd.reg_lfo_control = value;
}

void reset_sound()
{
	sound_samples = 0;
	sound_pos = 0;

	snd.last_cpu = 0;
	snd.reg_lfo_control = LFO_OFF;
	snd.reg_lfo_freq = 0;
	snd.reg_v = 0;
	snd.reg_vol = 0;

	for (int i = 0; i < 6; i++)
	{
		snd.reg_noise[i] = 0;
		snd.reg_status[i] = 0;
		snd.reg_balance[i] = 0;
		snd.reg_freq[i].W = 0;
		snd.reg_samples_pos[i] = 0;

		snd.change_samples[i] = 1;
		snd.ch_vol[i] = 0;
		snd.noise_delta[i] = 0;
		snd.noise_sample[i] = 0;
		snd.noise_sample_xor[i] = 0;
		snd.noise_random[i] = 0x12345678 + i;
		snd.noise_freq[i] = 0;

		snd.voice_delta[i] = 0;
		snd.voice_freq[i] = 0;
		snd.reg_voice_pos[i] = 0;
		snd.dda_sample[i] = 0;
		snd.reg_dda_value[i] = 0;

		snd.generator[i] = generator_silence;

		for (int j = 0; j < 32; j++) snd.reg_samples[i][j] = 0;
	}
}

void init_sound()
{
	reset_sound();
}
